import keras
from keras import ops
from keras.constraints import Constraint
from keras.initializers import Constant, Initializer
from keras.regularizers import Regularizer
from keras.src import backend
from keras.src.backend.config import epsilon
from quantizers.fixed_point._fixed_point_ops import _get_fixed_quantizer as get_fixed_quantizer
from quantizers.fixed_point._fixed_point_ops import round_conv

from ...constraints import MinMax
from .base import TrainableQuantizerBase, numbers


def minimal_i_given_xb(x, b, symmetric=False):
    eps = epsilon()
    if symmetric:
        return ops.ceil(ops.log2((x + eps) / (1 - 2.0**-b + eps)))
    i_pos = ops.ceil(ops.log2((x + eps) / (1 - 2.0**-b + eps)))
    i_neg = ops.ceil(ops.log2(-x - eps))
    return ops.where(b > 0, ops.where(x >= 0, i_pos, i_neg), 32)


def minimal_i_given_xf(x, f, symmetric=False):
    eps = epsilon()
    if symmetric:
        return ops.ceil(ops.log2(ops.abs(x) + 2**-f))
    i_pos = ops.ceil(ops.log2(x + 2**-f))
    i_neg = ops.ceil(ops.log2(-x - eps))
    return ops.where(x >= 0, i_pos, i_neg)


class FixedPointQuantizerBase(TrainableQuantizerBase):
    """Abstract base class for all fixed-point quantizers."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._seed_gen = None
        self._is_weight = False

    @property
    def round_mode(self) -> str:
        return self._round_mode

    @property
    def overflow_mode(self) -> str:
        return self._overflow_mode

    @property
    def seed_gen(self):
        return self._seed_gen

    def build(self, input_shape):
        super().build(input_shape)
        if self.round_mode.startswith('S_'):
            self._seed_gen = keras.random.SeedGenerator(self._seed)
        self.stateless_quantizer = get_fixed_quantizer(self.round_mode, self.overflow_mode)
        if self.overflow_mode == 'WRAP':
            init = Constant(self._i_decay_speed0) if not isinstance(self._i_decay_speed0, Initializer) else self._i_decay_speed0
            self._i_decay_speed = self.add_weight(name='i_decay_speed', shape=(), initializer=init, trainable=False)
        self._symmetric = self.overflow_mode.endswith('SYM')

    @property
    def symmetric(self):
        return self._symmetric

    @property
    def i_decay_speed(self):
        return ops.cast(self._i_decay_speed, self._i_decay_speed.dtype)

    @property
    def kif(self):
        raise NotImplementedError

    @property
    def k(self):
        raise NotImplementedError

    @property
    def b(self):
        raise NotImplementedError

    @property
    def i(self):
        raise NotImplementedError

    @property
    def f(self):
        raise NotImplementedError

    @property
    def bits(self):
        if self.overflow_mode != 'SAT':
            return self.b
        else:
            return self.b + self.k

    @property
    def fbits(self):
        b = self.b
        if self.overflow_mode == 'WRAP':
            return b + self.k * ops.cast(b > 0, self.dtype)
        else:
            return b + self.k

    @property
    def min(self):
        if self.symmetric:
            return -self.k * (2.0**self.i - 2.0**-self.f)
        else:
            return -self.k * (2.0**self.i)

    @property
    def max(self):
        return 2.0**self.i - 2.0**-self.f

    @property
    def epsilon(self):
        return 2.0**-self.f

    def get_any_k(self, inputs):
        return self.bw_mapper.x_to_bw_sign(inputs)

    def __repr__(self) -> str:
        if not self.built:
            return f'{self.__class__.__name__}({self.round_mode}, {self.overflow_mode}, name={self.name}, built=False)'
        k, i, f = self.k, self.i, self.f
        kstd, istd, fstd = float(ops.std(k)), float(ops.std(i)), float(ops.std(f))  # type: ignore
        kmean, imean, fmean = float(ops.mean(k)), float(ops.mean(i)), float(ops.mean(f))  # type: ignore
        kstr = f'{kmean:.2f}±{kstd:.2f}'
        istr = f'{imean:.2f}±{istd:.2f}'
        fstr = f'{fmean:.2f}±{fstd:.2f}'
        return (
            f'{self.__class__.__name__}(k={kstr}, i={istr}, f={fstr}, {self.round_mode}, {self.overflow_mode}, name={self.name})'
        )

    def get_minimum_i(self, inputs):
        raise NotImplementedError

    def call(self, inputs, training=None):  # type: ignore
        k, i, f = self.kif
        k = self.bw_mapper.bw_to_x(k, ops.shape(inputs))
        i = self.bw_mapper.bw_to_x(i, ops.shape(inputs))
        f = self.bw_mapper.bw_to_x(f, ops.shape(inputs))
        ret = self.stateless_quantizer(inputs, k, i, f, training is True, self.seed_gen)
        if not training:
            ret = ops.where(k + i + f > 0, ret, ops.zeros_like(ret))  # type: ignore
        return ret


class FixedPointQuantizerKBI(FixedPointQuantizerBase):
    """Internal quantizer for fixed-point quantization parameterized by keep_negative, bits, and integer bits.
    Can be used as a quantizer in Keras layers, but is usually wrapped by a `Quantizer` class to provide a consistent interface.

    Parameters
    ----------
    k0 : numbers | bool | Initializer
        Initial value for the keep_negative parameter. Not trained, but can be manually updated.
    b0 : numbers | Initializer
        Initial value for the number of bits. Trainable.
    i0 : numbers | Initializer
        Initial value for the number of integer bits. Trainable.
    round_mode : str
        Rounding mode, one of 'RND', 'TRN', 'RND_CONV', 'S_RND', 'S_RND_CONV'.
    overflow_mode : str
        Overflow mode, one of 'WRAP', 'SAT', 'SYM', 'SAT_SYM'.
    bc : Constraint | None
        Constraint for the number of bits.
    ic : Constraint | None
        Constraint for the number of integer bits.
    br : Regularizer | None
        Regularizer for the number of bits.
    ir : Regularizer | None
        Regularizer for the number of integer bits.
    i_decay_speed : numbers
        Speed of decay for the integer bits in WRAP mode, per step. If set to negative, enable tracing of maximum number even not in training mode.
    """

    def __init__(
        self,
        k0: numbers | bool | Initializer,
        b0: numbers | Initializer,
        i0: numbers | Initializer,
        round_mode: str,
        overflow_mode: str,
        bc: Constraint | None = MinMax(0, 12),
        ic: Constraint | None = None,
        br: Regularizer | None = None,
        ir: Regularizer | None = None,
        i_decay_speed: numbers = float('inf'),
        is_weight: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not isinstance(k0, Initializer) and not isinstance(b0, Initializer) and not isinstance(i0, Initializer):
            k0 = int(k0)
            assert k0 == 0 or k0 == 1, f'Invalid k0 value {k0}: must be 0 or 1.'
            assert b0 >= 0, f'Invalid b0 value {b0}: must be non-negative.'
        self._k0 = k0 if isinstance(k0, Initializer) else Constant(float(k0))
        self._b0 = b0 if isinstance(b0, Initializer) else Constant(float(b0))
        self._i0 = i0 if isinstance(i0, Initializer) else Constant(float(i0))
        self._i_decay_speed0 = i_decay_speed if isinstance(i_decay_speed, Initializer) else Constant(float(i_decay_speed))
        self._round_mode = round_mode.upper()
        self._overflow_mode = overflow_mode.upper()
        self.b_constraint = bc
        self.i_constraint = ic
        self.b_regularizer = br
        self.i_regularizer = ir
        self._is_weight = is_weight

        self.validate_config()

    def build(self, input_shape):
        bw_shape = self.bw_mapper.inference_weight_shape(input_shape)

        self._k = self.add_weight(
            name='k',
            shape=bw_shape,
            initializer=self._k0,
            trainable=False,
            dtype='uint8',
        )
        self._b = self.add_weight(
            name='b',
            shape=bw_shape,
            initializer=self._b0,
            trainable=True,
            constraint=self.b_constraint,
            regularizer=self.b_regularizer,
        )
        i_trainable = self.overflow_mode != 'WRAP'
        self._i = self.add_weight(
            name='i',
            shape=bw_shape,
            initializer=self._i0,
            trainable=i_trainable,
            constraint=self.i_constraint,
            regularizer=self.i_regularizer,
        )
        super().build(input_shape)

    @property
    def k(self):
        return backend.convert_to_tensor(self._k, self.dtype)

    @property
    def b(self):
        return round_conv(ops.cast(self._b, ops.dtype(self._b)))

    @property
    def i(self):
        return round_conv(ops.cast(self._i, ops.dtype(self._i)))

    @property
    def f(self):
        return self.b - self.i  # type: ignore

    @property
    def kif(self):
        k = self.k
        b = self.b
        i = self.i
        return k, i, b - i  # type: ignore

    def get_minimal_i(self, inputs):
        xr = self.bw_mapper.x_to_bw_absmax(inputs)
        return minimal_i_given_xb(xr, self.b, self.symmetric)

    def validate_config(self):
        assert self.b_constraint is None or isinstance(self.b_constraint, Constraint)
        assert self.i_constraint is None or isinstance(self.i_constraint, Constraint)
        assert self.b_regularizer is None or isinstance(self.b_regularizer, Regularizer)
        assert self.i_regularizer is None or isinstance(self.i_regularizer, Regularizer)

    def call(self, inputs, training=None):
        if self.overflow_mode == 'WRAP' and self.trainable:
            stochastic = self.stateless_quantizer.stochastic and training is True
            if training or training == 'tracing':  # noqa: E712, training maybe a special wrapper object
                new_i = self.get_minimal_i(inputs)
                if training:
                    new_i = ops.stop_gradient(ops.maximum((self._i - self.i_decay_speed), new_i))  # type: ignore
                    if self._i.constraint is not None:
                        new_i = self._i.constraint(new_i)
                    self._i.assign(new_i)
                    f = self.bw_mapper.bw_to_x(self.f, ops.shape(inputs))
                    return self.stateless_quantizer.round(inputs, f, stochastic, self.seed_gen)
                else:
                    new_i = ops.stop_gradient(ops.maximum(self.i, new_i))  # type: ignore
                    if self._i.constraint is not None:
                        new_i = self._i.constraint(new_i)

                    f = self.bw_mapper.bw_to_x(self.b - new_i, ops.shape(inputs))  # type: ignore
                    rinputs = self.stateless_quantizer.round(inputs, f, stochastic, self.seed_gen)
                    new_k = self.get_any_k(rinputs) & (self.b > 0)  # type: ignore
                    new_k = ops.cast(ops.cast(self.k, 'bool') | new_k, self._k.dtype)  # type: ignore
                    self._k.assign(new_k)
                    self._i.assign(new_i)
                    return rinputs

            if self._is_weight:
                f = self.bw_mapper.bw_to_x(self.f, ops.shape(inputs))
                if self.stateless_quantizer.round_mode == 'RND':
                    return self.stateless_quantizer.round(inputs, f, False, self.seed_gen)

        return super().call(inputs, training is True)


class FixedPointQuantizerKIF(FixedPointQuantizerBase):
    """Internal quantizer for fixed-point quantization parameterized by keep_negative, integer bits, and fractional bits.
    Can be used as a quantizer in Keras layers, but is usually wrapped by a `Quantizer` class to provide a consistent interface.

    Parameters
    ----------
    k0 : numbers | bool | Initializer
        Initial value for the keep_negative parameter. Not trained, but can be manually updated.
    i0 : numbers | Initializer
        Initial value for the number of integer bits. Trainable.
    f0 : numbers | Initializer
        Initial value for the number of fractional bits. Trainable.
    round_mode : str
        Rounding mode, one of 'RND', 'TRN', 'RND_CONV', 'S_RND', 'S_RND_CONV'.
    overflow_mode : str
        Overflow mode, one of 'WRAP', 'SAT', 'SYM', 'SAT_SYM'.
    ic : Constraint | None
        Constraint for the number of integer bits.
    fc : Constraint | None
        Constraint for the number of fractional bits.
    ir : Regularizer | None
        Regularizer for the number of integer bits.
    fr : Regularizer | None
        Regularizer for the number of fractional bits.
    i_decay_speed : numbers
        Speed of decay for the integer bits in WRAP mode, per step. If set to negative, enable tracing of maximum number even not in training mode.
    """

    def __init__(
        self,
        k0: numbers | bool,
        i0: numbers,
        f0: numbers,
        round_mode: str,
        overflow_mode: str,
        ic: Constraint | None = None,
        fc: Constraint | None = None,
        ir: Regularizer | None = None,
        fr: Regularizer | None = None,
        i_decay_speed: numbers = float('inf'),
        is_weight: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not isinstance(k0, Initializer) and not isinstance(i0, Initializer) and not isinstance(f0, Initializer):
            k0 = int(k0)
            assert k0 == 0 or k0 == 1, f'Invalid k0 value {k0}: must be 0 or 1.'
            assert i0 + f0 >= 0, f'Invalid i0+f0 value {i0 + f0}: must be non-negative.'
        self._k0 = k0 if isinstance(k0, Initializer) else Constant(float(k0))
        self._i0 = i0 if isinstance(i0, Initializer) else Constant(float(i0))
        self._f0 = f0 if isinstance(f0, Initializer) else Constant(float(f0))
        self._i_decay_speed0 = i_decay_speed if isinstance(i_decay_speed, Initializer) else Constant(float(i_decay_speed))
        self._round_mode = round_mode.upper()
        self._overflow_mode = overflow_mode.upper()
        self.i_constraint = ic
        self.f_constraint = fc
        self.i_regularizer = ir
        self.f_regularizer = fr
        self._is_weight = is_weight
        self.validate_config()

    def build(self, input_shape):
        bw_shape = self.bw_mapper.inference_weight_shape(input_shape)

        self._k = self.add_weight(
            name='k',
            shape=bw_shape,
            initializer=self._k0,
            trainable=False,
            dtype='uint8',
        )
        i_trainable = self.overflow_mode != 'WRAP'
        self._i = self.add_weight(
            name='i',
            shape=bw_shape,
            initializer=self._i0,
            trainable=i_trainable,
            constraint=self.i_constraint,
            regularizer=self.i_regularizer,
        )
        self._f = self.add_weight(
            name='f',
            shape=bw_shape,
            initializer=self._f0,
            trainable=True,
            constraint=self.f_constraint,
            regularizer=self.f_regularizer,
        )

        super().build(input_shape)

    def call(self, inputs, training=None):  # type: ignore
        if self.overflow_mode == 'WRAP' and self.trainable:
            f = self.bw_mapper.bw_to_x(self.f, ops.shape(inputs))
            stochastic = self.stateless_quantizer.stochastic and training is True
            rinputs = self.stateless_quantizer.round(inputs, f, stochastic, self.seed_gen)

            if training or training == 'tracing':  # noqa: E712, training maybe a special wrapper object
                _new_i = self.get_minimal_i(rinputs)
                if training:
                    new_i = ops.stop_gradient(ops.maximum(self._i - self.i_decay_speed, _new_i))  # type: ignore
                    if self._i.constraint is not None:
                        _new_i = self._i.constraint(_new_i)
                    self._i.assign(new_i)
                else:
                    # tracing
                    if self._i.constraint is not None:
                        _new_i = self._i.constraint(_new_i)
                    new_i = ops.stop_gradient(ops.maximum(self.i, _new_i))  # type: ignore
                    new_k = self.get_any_k(rinputs)
                    new_k = ops.cast(ops.cast(self.k, 'bool') | new_k, self._k.dtype)  # type: ignore
                    self._k.assign(new_k)
                    self._i.assign(new_i)

                return rinputs

            if self._is_weight:
                return rinputs

        return super().call(inputs, training is True)

    @property
    def k(self):
        return ops.cast(self._k, self.dtype)

    @property
    def b(self):
        return ops.relu(self.i + self.f)  # type: ignore

    @property
    def i(self):
        return round_conv(ops.cast(self._i, self._i.dtype))

    @property
    def f(self):
        return round_conv(ops.cast(self._f, self._f.dtype))

    @property
    def kif(self):
        return self.k, self.i, self.f

    def get_minimal_i(self, inputs):
        xr = self.bw_mapper.x_to_bw_absmax(inputs)
        return minimal_i_given_xf(xr, self.f, self.symmetric)

    def validate_config(self):
        assert self.f_constraint is None or isinstance(self.f_constraint, Constraint)
        assert self.i_constraint is None or isinstance(self.i_constraint, Constraint)
        assert self.f_regularizer is None or isinstance(self.f_regularizer, Regularizer)
        assert self.i_regularizer is None or isinstance(self.i_regularizer, Regularizer)



from keras import ops
from keras.regularizers import Regularizer
from keras.saving import register_keras_serializable

from ..utils.misc import numbers


@register_keras_serializable(package='hgq')
class MonoL1(Regularizer):
    def __init__(self, l1: numbers):
        self.l1 = float(l1)

    def __call__(self, x):
        return self.l1 * ops.sum(x)  # type: ignore

    def get_config(self):
        return {'l1': self.l1}



from typing import TypeVar

import numpy as np
from da4ml.trace import FixedVariableArray
from keras import ops
from keras.src.ops.image import ExtractPatches, extract_patches_3d

from hgq.layers import QConv1D, QConv2D, QConv3D

from ._base import ReplayOperationBase, to_np_arr

T = TypeVar('T', FixedVariableArray, np.ndarray)


def symbolic_extract_patches_3d(
    images: T,
    size: tuple[int, int, int],
    strides: tuple[int, int, int],
    dilation_rate: tuple[int, int, int],
    padding: str,
    data_format: str,
    pad_value: float = 0,
) -> T:
    img_tensor = ops.cast(ops.reshape(ops.arange(images.size), images.shape), dtype='float32')
    img_tensor = -img_tensor - 1
    out_tensor = extract_patches_3d(
        img_tensor[None],
        size=size,
        strides=strides,
        dilation_rate=dilation_rate,  # type: ignore
        padding=padding,
        data_format=data_format,
    )[0]
    out_index: np.ndarray = ops.convert_to_numpy(out_tensor).round().astype(np.int32)  # type: ignore
    mask = out_index == 0
    out_index = np.where(mask, 0, -out_index - 1)
    images = images.ravel()[out_index]

    if isinstance(images, FixedVariableArray):
        _vars = images._vars
        _vars = np.where(mask, pad_value, _vars)
        images = FixedVariableArray(_vars, images.solver_options)  # type: ignore
    else:
        images = np.where(mask, pad_value, images)

    return images


def symbolic_extract_patches(
    images: T,
    size: tuple[int, ...] | int,
    strides: tuple[int, ...] | int,
    dilation_rate: tuple[int, ...] | int,
    padding: str,
    data_format: str,
    pad_value: float = 0,
) -> T:
    rank = images.ndim - 1
    size = (size,) * rank if isinstance(size, int) else size
    strides = (strides,) * rank if isinstance(strides, int) else strides
    dilation_rate = (dilation_rate,) * rank if isinstance(dilation_rate, int) else dilation_rate

    assert rank == len(size) == len(strides) == len(dilation_rate), (
        f'Invalid rank {rank} for size {size}, strides {strides}, dilation_rate {dilation_rate}'
    )

    pad_rank = 3 - rank
    _size: tuple[int, int, int] = (1,) * pad_rank + size  # type: ignore
    _strides: tuple[int, int, int] = (1,) * pad_rank + strides  # type: ignore
    _dilation_rate: tuple[int, int, int] = (1,) * pad_rank + dilation_rate  # type: ignore

    _pad = (1,) * pad_rank
    if data_format == 'channels_first':
        images = np.moveaxis(images, 0, -1)  # type: ignore

    *spa, ch = images.shape
    images = images.reshape(*_pad, *spa, ch)

    r = symbolic_extract_patches_3d(
        images,
        size=_size,
        strides=_strides,
        dilation_rate=_dilation_rate,
        padding=padding,
        data_format='channels_last',
        pad_value=pad_value,
    )

    return r.reshape(r.shape[pad_rank:])


class ReplayExtractPatches(ReplayOperationBase):
    handles = (ExtractPatches,)

    def call(self, images: FixedVariableArray) -> FixedVariableArray:
        op: ExtractPatches = self.op
        pixel_shape = op.size
        strides = op.strides
        dilation_rate: int | tuple[int, int] = op.dilation_rate
        padding = op.padding
        data_format = op.data_format

        if strides is None:
            strides = 1

        return symbolic_extract_patches(images, pixel_shape, strides, dilation_rate, padding, data_format)


class ReplayQConv(ReplayOperationBase):
    handles = (QConv1D, QConv2D, QConv3D)

    def call(self, inputs: FixedVariableArray) -> FixedVariableArray:
        layer: QConv1D | QConv2D | QConv3D = self.op
        qkernel = to_np_arr(layer.qkernel)
        qbias = to_np_arr(layer.qbias) if layer.qbias is not None else None
        strides = layer.strides
        padding = layer.padding
        dilation_rate = layer.dilation_rate
        groups = layer.groups

        x = symbolic_extract_patches(
            inputs,
            size=layer.kernel_size,
            strides=strides,
            dilation_rate=dilation_rate,
            padding=padding,
            data_format=layer.data_format,
        )

        if layer.data_format == 'channels_first':
            inputs = np.moveaxis(inputs, 0, -1)  # type: ignore

        ch_out = qkernel.shape[-1]

        _ch_out = ch_out // groups

        x = x.reshape(*x.shape[:-1], -1, groups)
        kernel = qkernel.reshape(-1, groups, _ch_out)

        outputs = np.einsum('...ig,igo->...go', x, kernel)  # type: ignore
        outputs = outputs.reshape(*outputs.shape[:-2], -1) + qbias

        if layer.data_format == 'channels_first':
            outputs: FixedVariableArray = np.moveaxis(outputs, -1, 0)  # type: ignore

        return outputs



from .layer import LayerConfigScope, global_config
from .quantizer import HardSigmoidConfig, HardTanhConfig, QuantizerConfig, QuantizerConfigScope, default_configs

__all__ = [
    'LayerConfigScope',
    'QuantizerConfigScope',
    'QuantizerConfig',
    'global_config',
    'default_configs',
    'HardTanhConfig',
    'HardSigmoidConfig',
]



from collections.abc import Callable

from keras import ops
from keras.layers import Activation
from keras.src import backend

from ..quantizer import QuantizerConfig
from .core.base import QLayerBaseSingleInput


def _large_negative_number(dtype):
    """Return a Large negative number based on dtype."""
    if backend.standardize_dtype(dtype) == 'float16':
        return -3e4
    return -1e9


class QUnaryFunctionLUT(Activation, QLayerBaseSingleInput):
    def __init__(
        self,
        activation: Callable | str,
        iq_conf: QuantizerConfig | None = None,
        oq_conf: QuantizerConfig | None = None,
        enable_oq=True,
        enable_iq=True,
        allow_heterogeneous_input: bool = True,
        allow_heterogeneous_table: bool = True,
        override_oq_k0_to_0: bool = False,
        **kwargs,
    ):
        act_name = activation.__name__ if isinstance(activation, Callable) else activation
        assert act_name not in ('softmax', 'log_softmax'), f'activation {act_name} is not unary'

        self._allow_heterogeneous_table = allow_heterogeneous_table
        self._allow_heterogeneous_input = allow_heterogeneous_input

        if enable_oq:
            oq_conf = oq_conf or QuantizerConfig('default', 'table')
            if override_oq_k0_to_0:
                if 'k0' in oq_conf.config:
                    oq_conf.config['k0'] = False
            if not allow_heterogeneous_table:
                oq_conf.config['homogeneous_axis'] = None
                oq_conf.config['heterogeneous_axis'] = ()
        else:
            self._enable_ebops = False

        if enable_iq and not allow_heterogeneous_input:
            iq_conf = iq_conf or QuantizerConfig('default', 'datalane')
            iq_conf.config['homogeneous_axis'] = None
            iq_conf.config['heterogeneous_axis'] = ()

        super().__init__(
            activation=activation, iq_conf=iq_conf, oq_conf=oq_conf, enable_oq=enable_oq, enable_iq=enable_iq, **kwargs
        )
        self.built = False

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        return self.activation(inputs)

    def _compute_ebops(self, shape):
        bw_inp = self.iq.bits_(shape)
        bw_out = self.oq.bits_(shape)
        # TODO: more realistic cost for lookup tables
        return ops.sum((2.0**bw_inp) * bw_out) * 1e-4  # type: ignore

    def get_config(self):
        config = super().get_config()
        config['allow_heterogeneous_table'] = self._allow_heterogeneous_table
        config['allow_heterogeneous_input'] = self._allow_heterogeneous_input
        return config


class QAffinedUnaryFunctionLUT(QUnaryFunctionLUT):
    def __init__(
        self,
        activation: Callable | str,
        iq_conf: QuantizerConfig | None = None,
        oq_conf: QuantizerConfig | None = None,
        enable_oq=True,
        enable_iq=True,
        allow_heterogeneous_input: bool = True,
        allow_heterogeneous_table: bool = True,
        **kwargs,
    ):
        super().__init__(
            activation=activation,
            iq_conf=iq_conf,
            oq_conf=oq_conf,
            enable_oq=enable_oq,
            enable_iq=enable_iq,
            allow_heterogeneous_input=allow_heterogeneous_input,
            allow_heterogeneous_table=allow_heterogeneous_table,
            **kwargs,
        )
        self.built = False

    def build(self, input_shape):
        self.scale = self.add_weight(name='scale', shape=(1,), initializer='ones', trainable=True)
        self.bias = self.add_weight(name='bias', shape=(1,), initializer='zeros', trainable=True)
        super().build(input_shape)

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        inputs = inputs * self.scale + self.bias
        return self.activation(inputs)



import keras
from da4ml.trace import FixedVariableArray
from da4ml.trace.ops import einsum

from hgq.layers import (
    QBatchNormDense,
    QDense,
    QEinsumDense,
    QEinsumDenseBatchnorm,
)

from ._base import ReplayOperationBase, to_np_arr


class ReplayQDense(ReplayOperationBase):
    handles = (QDense, QEinsumDense, QEinsumDenseBatchnorm, QBatchNormDense, keras.layers.EinsumDense)

    def call(self, inputs: FixedVariableArray) -> FixedVariableArray:
        op = self.op
        if isinstance(op, (QDense, QBatchNormDense)):
            qkernel = op.qkernel
            qbias = op.qbias
            eq = '...c,cC->...C'
        elif isinstance(op, (QEinsumDense, QEinsumDenseBatchnorm)):
            qkernel = op.qkernel
            qbias = op.qbias
            eq = op.equation
        elif isinstance(op, keras.layers.EinsumDense):
            qkernel = op.kernel
            qbias = op.bias
            eq = op.equation
        else:
            raise TypeError(f'Unsupported layer type: {type(op)}')

        qkernel = to_np_arr(qkernel)
        qbias = to_np_arr(qbias) if qbias is not None else None
        return (einsum(eq, inputs[None], qkernel) + qbias)[0]


__all__ = ['ReplayQDense']



import numpy as np
import pytest
from keras import Model, ops

from hgq.config import QuantizerConfigScope
from hgq.layers import QAffinedUnaryFunctionLUT, QSoftmax, QUnaryFunctionLUT

from .base import LayerTestBase


def custom_fn(x):
    return ops.tanh(x) - ops.sin(x) + ops.log(ops.abs(x) + 1)  # type: ignore


class TestQUnaryFunctionLUT(LayerTestBase):
    layer_cls = QUnaryFunctionLUT
    custom_objects = {'custom_fn': custom_fn}

    @pytest.fixture(params=['sigmoid', 'tanh', custom_fn])
    def layer_kwargs(self, request):
        return {'activation': request.param, 'allow_heterogeneous_table': False}

    @pytest.fixture
    def input_shapes(self):
        return (8,)


class TestQAffinedUnaryFunctionLUT(TestQUnaryFunctionLUT):
    hls4ml_not_supported = True
    layer_cls = QAffinedUnaryFunctionLUT


class TestSoftmax(LayerTestBase):
    layer_cls = QSoftmax

    @pytest.fixture(params=[-1, (-2, -1), -2])
    def axis(self, request):  # return request.param
        return request.param

    @pytest.fixture(params=[(8,), (4, 4)])
    def input_shapes(self, request, axis):
        if axis != -1 and len(request.param) == 1:
            pytest.xfail('Invalid axis config for shape')
        return request.param

    @pytest.fixture(params=[True, False])
    def stable(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, axis, stable, use_parallel_io):
        return {
            'axis': axis,
            'stable': stable,
            'enable_iq': True,
        }

    def test_behavior(self, input_data, layer_kwargs):
        with QuantizerConfigScope(default_q_type='dummy'):
            softmax = QSoftmax(**layer_kwargs)

        axis = layer_kwargs['axis']
        hgq_output = softmax(input_data)
        if not isinstance(axis, tuple):
            ref_output = ops.nn.softmax(input_data, axis=axis)
        else:
            shape = input_data.shape
            input_data = ops.reshape(input_data, shape[:-2] + (-1,))
            ref_output = ops.reshape(ops.nn.softmax(input_data, axis=-1), shape)

        hgq_output_np: np.ndarray = ops.convert_to_numpy(hgq_output)  # type: ignore
        ref_output_np: np.ndarray = ops.convert_to_numpy(ref_output)  # type: ignore

        np.testing.assert_allclose(hgq_output_np, ref_output_np, atol=1e-6)

    def test_hls4ml_conversion(  # type: ignore
        self, model: Model, input_data, temp_directory: str, use_parallel_io: bool, q_type: str, axis: int
    ):
        if not use_parallel_io and axis != -1:
            pytest.skip('hls4ml only support axis=-1 with io_stream')
        return super().test_hls4ml_conversion(model, input_data, temp_directory, use_parallel_io, q_type)



from ._base import _registry
from .activation import *
from .attn import *
from .batchnorm import *
from .conv import *
from .dense import *
from .ops import *
from .pool import *
from .table import *

__all__ = ['_registry']



import keras
import numpy as np
import pytest

from hgq.layers import QConv1D, QConv2D

from .base import LayerTestBase


class TestConv1D(LayerTestBase):
    layer_cls = QConv1D

    @pytest.fixture(params=[1, 4])
    def ch_out(self, request):
        return request.param

    @pytest.fixture(
        params=[
            {'kernel_size': 1, 'strides': 2, 'parallelization_factor': -1},
            {'kernel_size': 3, 'strides': 1, 'parallelization_factor': -1},
            {'kernel_size': 2, 'strides': 2, 'parallelization_factor': 1},
        ]
    )
    def conv_params(self, request):
        return request.param

    @pytest.fixture()
    def input_shapes(self):
        return (6, 2)

    @pytest.fixture(params=['valid', 'same'])
    def padding(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, ch_out, conv_params):
        return {'filters': ch_out, **conv_params}

    def assert_equal(self, keras_output, hls_output):
        if keras.backend.backend() == 'torch':
            # Torch conv operator introduces some extra numerical error
            return np.testing.assert_allclose(keras_output, hls_output, atol=1e-4)
        return np.testing.assert_allclose(keras_output, hls_output, atol=1e-6)


class TestConv2D(LayerTestBase):
    layer_cls = QConv2D

    @pytest.fixture(params=[1, 4])
    def ch_out(self, request):
        return request.param

    @pytest.fixture(
        params=[
            {'kernel_size': (1, 1), 'strides': (3, 5), 'parallelization_factor': -1},
            {'kernel_size': (3, 3), 'strides': (1, 1), 'parallelization_factor': -1},
            {'kernel_size': (3, 2), 'strides': (1, 3), 'parallelization_factor': 1},
            {'kernel_size': (2, 3), 'strides': (4, 2), 'parallelization_factor': 1},
        ]
    )
    def conv_params(self, request):
        return request.param

    @pytest.fixture()
    def input_shapes(self):
        return (6, 6, 2)

    @pytest.fixture(params=['valid', 'same'])
    def padding(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, ch_out, conv_params):
        return {'filters': ch_out, **conv_params}

    def assert_equal(self, keras_output, hls_output):
        # Conv operations has some numerical error in ops.conv operations.
        if keras.backend.backend() == 'torch':
            # Torch conv operator introduces some extra numerical error
            return np.testing.assert_allclose(keras_output, hls_output, atol=5e-4)
        return np.testing.assert_allclose(keras_output, hls_output, atol=1e-6)

    def test_training(self, model: keras.Model, input_data: np.ndarray, overflow_mode, ch_out: int):
        if keras.backend.backend() == 'torch' and ch_out == 1:
            pytest.skip('Torch runtime error for unknown reason when ch_out is 1.')
        return super().test_training(model, input_data, overflow_mode)



from keras import ops
from keras.constraints import Constraint
from keras.saving import register_keras_serializable
from keras.src import backend

from ..utils.misc import numbers

__all__ = ['MinMax', 'Min', 'Max', 'Constant']


@register_keras_serializable(package='hgq')
class MinMax(Constraint):
    """Constrains the weights to between min_value and max_value."""

    def __init__(self, min_value: numbers, max_value: numbers):
        self.min_value = float(min_value)
        self.max_value = float(max_value)

    def __call__(self, w):
        w = backend.convert_to_tensor(w)
        return ops.clip(w, self.min_value, self.max_value)

    def get_config(self):
        return {'min_value': self.min_value, 'max_value': self.max_value}


@register_keras_serializable(package='hgq')
class Constant(Constraint):
    """Constrains the weights to a constant value."""

    def __init__(self, value: numbers):
        self.value = float(value)

    def __call__(self, w):
        w = backend.convert_to_tensor(w)
        return ops.full_like(w, self.value)

    def get_config(self):
        return {'value': self.value}


@register_keras_serializable(package='hgq')
class Min(Constraint):
    """Constrains the weights to greater or equal than min_value."""

    def __init__(self, min_value: numbers):
        self.min_value = float(min_value)

    def __call__(self, w):
        w = backend.convert_to_tensor(w)
        return ops.maximum(w, self.min_value)

    def get_config(self):
        return {'min_value': self.min_value}


@register_keras_serializable(package='hgq')
class Max(Constraint):
    """Constrains the weights to less or equal than max_value."""

    def __init__(self, max_value: numbers):
        self.max_value = float(max_value)

    def __call__(self, w):
        w = backend.convert_to_tensor(w)
        return ops.minimum(w, self.max_value)

    def get_config(self):
        return {'max_value': self.max_value}



import keras
import numpy as np
from da4ml.trace import FixedVariableArray
from da4ml.trace.ops import relu
from keras.layers import LeakyReLU, PReLU, ReLU

from hgq.layers import (
    QAffinedUnaryFunctionLUT,
    QSoftmax,
    QUnaryFunctionLUT,
)

from ._base import ReplayOperationBase, to_np_arr


class ReplayReLU(ReplayOperationBase):
    handles = (ReLU, LeakyReLU, PReLU)

    def call(self, inputs: FixedVariableArray) -> FixedVariableArray:
        op = self.op
        if isinstance(op, ReLU):
            th, neg, maxv = op.threshold, op.negative_slope, op.max_value
        elif isinstance(op, LeakyReLU):
            th, neg, maxv = 0, op.negative_slope, None
        elif isinstance(op, PReLU):
            th, neg, maxv = 0, to_np_arr(op.alpha), None
        else:
            raise TypeError(f'Unsupported activation layer: {type(op)}')

        if th == 0 and np.all(neg == 0) and maxv is None:
            return relu(inputs)

        pos_part = inputs if maxv is None else np.minimum(inputs, maxv)  # type: ignore
        pos_part = pos_part._vars.ravel()

        if th != 0:
            z_cond = (inputs - (th + 2.0 ** (-inputs.kif[2] - 1)))._vars.ravel()
        else:
            z_cond = inputs._vars.ravel()

        neg_part = ((inputs[None] - th) * neg)._vars.ravel()
        out = np.array([c.msb_mux(n, p) if c.low < 0 else p for c, n, p in zip(z_cond, neg_part, pos_part)])

        return FixedVariableArray(out.reshape(inputs.shape), inputs.solver_options)


class ReplayQFunctionLUT(ReplayOperationBase):
    __activation_handled__ = True
    handles = (QUnaryFunctionLUT, QAffinedUnaryFunctionLUT)

    def call(self, x: FixedVariableArray) -> FixedVariableArray:
        op: QUnaryFunctionLUT = self.op

        def activation(x) -> np.ndarray:
            kx = keras.ops.convert_to_tensor(x[None])
            if isinstance(op, QAffinedUnaryFunctionLUT):
                kx = kx * op.scale + op.bias
            kx = op.activation(kx)
            return keras.ops.convert_to_numpy(kx[0])  # type: ignore

        return x.apply(activation)


class ReplayQSoftmax(ReplayOperationBase):
    handles = (QSoftmax,)

    def call(self, inputs: FixedVariableArray, mask: None | FixedVariableArray = None) -> FixedVariableArray:
        op: QSoftmax = self.op
        inputs = inputs[None]

        if op.stable:
            if mask is not None:
                low = np.min(inputs.lhs[0]) - 1
                inputs = np.where(mask, inputs, low)  # type: ignore
            inputs = np.amax(inputs, axis=op.axes, keepdims=True) - inputs  # type: ignore

        exp_inp = ReplayQFunctionLUT(op.exp_table)(inputs[0])[0]

        if mask is not None:
            exp_inp = mask * exp_inp

        sums = np.sum(exp_inp[None], axis=op.axes, keepdims=True)[0]  # type: ignore
        divisor = ReplayQFunctionLUT(op.inv_table)(sums)[0]

        return exp_inp * divisor


__all__ = ['ReplayReLU', 'ReplayQFunctionLUT', 'ReplayQSoftmax']
